//+------------------------------------------------------------------+
//|                                           ConstFileUnlimited.mq5 |
//+------------------------------------------------------------------+

#property library
#property copyright "TheXpert"
#property link      "theforexpert@gmail.com"
#property version   "1.00"

#include <TheXpert/FileUnlimited/Interfaces.mqh>
#include <TheXpert/StringUtils.mqh>

#include <TheXpert/DebugTools/ASSERT.mqh>
#include <TheXpert/DebugTools/Trace/Trace.mqh>

#import "kernel32.dll"
   bool   CloseHandle(int hObject);
   int    CreateFileW(     string lpFileName,         uint dwDesiredAccess, 
                           uint dwShareMode,          uint lpSecurityAttributes,
                           uint dwCreationDisposition,uint dwFlagsAndAttributes, 
                           int hTemplateFile
                     );
   uint   GetFileAttributesW(string lpFileName);
   uint   GetFileSize(int hFile, uint& lpFileSizeHigh[]);
   bool   GetFileTime(int hFile, uint& lpCreationTime[], uint& lpLastAccessTime[], uint& lpLastWriteTime[]);
   bool   ReadFile(int hFile, uchar& lpBuffer[], uint nNumberOfBytesToRead, uint& lpNumberOfBytesRead[], int lpOverlapped);
   uint   SetFilePointer(int hFile, int lDistanceToMove, int& lpDistanceToMoveHigh[], uint dwMoveMethod);
   bool   FileTimeToSystemTime(const uint& lpFileTime[], MqlDateTime& lpSystemTime[]);
#import

//========================================================


class ConstFileImpl
   : public ConstFile
{
public:
   ConstFileImpl(string path, string separators);
   
   bool IsValid() const 
   {
      return m_hFile != INVALID_HANDLE;
   }
   
   ~ConstFileImpl()
   {
      if (m_hFile != INVALID_HANDLE)
      {
         CloseHandle(m_hFile);
      }
   }
   
   virtual ulong Size() const 
   {
      return m_Size;
   }
   
   virtual string Path() const
   {
      return m_Path;
   }

   virtual datetime CreationTime() const 
   {
      return m_CreationTime;
   }
   
   virtual datetime LastAccessTime() const 
   {
      return m_LastAccessTime;
   }
   
   virtual datetime LastModifyTime() const 
   {
      return m_LastModifyTime;
   }
   
   virtual bool Read(bool& value, string separators);
   virtual bool Read(long& value, string separators);
   virtual bool Read(int& value, string separators);
   virtual bool Read(double& value, string separators);
   virtual bool Read(string& value, string separators);
   virtual bool Read(datetime& value, string separators);
   virtual bool Read(color& value, string separators);

   virtual bool Read(bool& value);
   virtual bool Read(long& value);
   virtual bool Read(int& value);
   virtual bool Read(double& value);
   virtual bool Read(string& value);
   virtual bool Read(datetime& value);
   virtual bool Read(color& value);

   virtual bool SetPos(long pos, MovePointerMethod from = MPM_BEGIN);
   virtual string ToString();
   
   virtual long Find(string what);
   
   virtual bool IsEnding() const;

private:
   void Init();
   bool ReadBlock(uint size = 1000);
   
   bool ReadString(string& value, string separators);
   
   void FixEndOfLine();

private:
   int m_hFile;
   ulong m_Size;
   datetime m_CreationTime;
   datetime m_LastAccessTime;
   datetime m_LastModifyTime;
   
   string m_Buffer;
   ulong m_Pos;
   uint m_Attributes;
   string m_Path;
   string m_Separators;
};

ConstFileImpl::ConstFileImpl(string path, string separators)
   : ConstFile(path)
   , m_hFile(INVALID_HANDLE)
   , m_Size(0)
   , m_CreationTime(0)
   , m_LastAccessTime(0)
   , m_LastModifyTime(0)
   , m_Buffer("")
   , m_Pos(0)
   , m_Attributes(INVALID_FILE_ATTRIBUTES)
   , m_Path(path)
   , m_Separators(separators)
{__TRACE
   m_hFile = CreateFileW( path, FU_GENERIC_READ, FU_FILE_SHARE_READ, 0, FU_OPEN_EXISTING, 0, 0);
   
   ASSERT(m_hFile != INVALID_HANDLE);
   if (m_hFile != INVALID_HANDLE)
   {
      Init();
   }
}

void ConstFileImpl::Init()
{__TRACE
   // requesting size
   uint hi[1] = {0};
   uint low = GetFileSize(m_hFile, hi);
   
   m_Size = ulong(hi);
   m_Size = m_Size >> 32;
   m_Size += low;
   
   // requesting time properties
   uint creation[4], lastAccess[4], lastModify[4];
   if (GetFileTime(m_hFile, creation, lastAccess, lastModify))
   {
      // request succeeded, need to tranform time
      MqlDateTime time[1];
      FileTimeToSystemTime(creation, time);
      m_CreationTime = StructToTime(time[0]);

      FileTimeToSystemTime(lastAccess, time);
      m_LastAccessTime = StructToTime(time[0]);

      FileTimeToSystemTime(lastModify, time);
      m_LastModifyTime = StructToTime(time[0]);
   }
   else ASSERT(false);
   
   m_Attributes = GetFileAttributesW(m_Path);
}

bool ConstFileImpl::Read(bool& value)
{
   return Read(value, m_Separators);
}

bool ConstFileImpl::Read(long& value)
{
   return Read(value, m_Separators);
}

bool ConstFileImpl::Read(int& value)
{
   return Read(value, m_Separators);
}

bool ConstFileImpl::Read(double& value)
{
   return Read(value, m_Separators);
}

bool ConstFileImpl::Read(string& value)
{
   return Read(value, m_Separators);
}

bool ConstFileImpl::Read(datetime& value)
{
   return Read(value, m_Separators);
}

bool ConstFileImpl::Read(color& value)
{
   return Read(value, m_Separators);
}

bool ConstFileImpl::Read(bool& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = StringToBool(tmp);
      return true;
   }
   return false;
}

bool ConstFileImpl::Read(long& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = StringToInteger(tmp);
      return true;
   }
   return false;
}

bool ConstFileImpl::Read(int& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = int(StringToInteger(tmp));
      return true;
   }
   return false;
}

bool ConstFileImpl::Read(double& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = StringToDouble(tmp);
      return true;
   }
   return false;
}

bool ConstFileImpl::ReadString(string& value, string separators)
{__TRACE
   if (m_Buffer == "")
   {
      if (!ReadBlock()) return false;
   }
   
   int pos = StringFindAnyOfChar(m_Buffer, separators);
   while (pos == 0)
   {
      m_Buffer = StringSubstr(m_Buffer, 1);
      m_Pos += 1;
      pos = StringFindAnyOfChar(m_Buffer, separators);
   }
   
   string strValue = "";
   
   bool needFixEOL = false;
   
   if (pos == -1)
   {
      strValue = m_Buffer;
      m_Pos += StringLen(m_Buffer);
      m_Buffer = "";
      
      bool succeeded = ReadBlock();
      
      while (succeeded)
      {
         if (strValue == "")
         {
            pos = StringFindAnyOfChar(m_Buffer, separators);
            while (pos = 0)
            {
               m_Buffer = StringSubstr(m_Buffer, 1);
               m_Pos += 1;
            }
            
            if (pos == -1)
            {
               strValue = m_Buffer;
               m_Pos += 1*StringLen(m_Buffer);
               m_Buffer = "";
            }
            else
            {
               strValue = StringSubstr(m_Buffer, 0, pos);

               ushort chr = StringGetCharacter(m_Buffer, pos);
               if (chr == '\r') needFixEOL = true;

               m_Pos += (StringLen(strValue) + 1);
               m_Buffer = StringSubstr(m_Buffer, pos + 1);
               break;
            }
         }
         else
         {
            pos = StringFindAnyOfChar(m_Buffer, separators);
            if (pos == 0)
            {
               ushort chr = StringGetCharacter(m_Buffer, pos);
               if (chr == '\r') needFixEOL = true;

               m_Buffer = StringSubstr(m_Buffer, 1);
               m_Pos += 1;
               break;
            }
            else if (pos == -1)
            {
               strValue = m_Buffer;
               m_Pos += StringLen(m_Buffer);
               m_Buffer = "";
            }
            else
            {
               strValue = strValue + StringSubstr(m_Buffer, 0, pos);

               ushort chr = StringGetCharacter(m_Buffer, pos);
               if (chr == '\r') needFixEOL = true;

               m_Pos += (StringLen(strValue) + 1);
               m_Buffer = StringSubstr(m_Buffer, pos + 1);
               break;
            }
         }
         succeeded = ReadBlock();
      }

      if (strValue == "") return false;
   }
   else
   {
      strValue = StringSubstr(m_Buffer, 0, pos);

      ushort chr = StringGetCharacter(m_Buffer, pos);
      if (chr == '\r') needFixEOL = true;

      m_Pos += (StringLen(strValue) + 1);
      m_Buffer = StringSubstr(m_Buffer, pos + 1);
   }
   
   if (needFixEOL) FixEndOfLine();
   
   value = strValue;
   
   return true;
}

bool ConstFileImpl::Read(string& value, string separators)
{__TRACE
   if (m_Buffer == "")
   {
      if (!ReadBlock()) return false;
   }
   
   bool needFixEOL = false;
   
   int pos = StringFindAnyOfChar(m_Buffer, separators);
   if (pos == 0)
   {
      ushort chr = StringGetCharacter(m_Buffer, pos);
      if (chr == '\r') needFixEOL = true;

      m_Buffer = StringSubstr(m_Buffer, 1);
      m_Pos += 1;
      value = "";
      
      if (needFixEOL) FixEndOfLine();
      
      return true;
   }
   
   string strValue = "";
   
   if (pos == -1)
   {
      strValue = m_Buffer;
      m_Pos += StringLen(m_Buffer);
      m_Buffer = "";
      
      bool succeeded = ReadBlock();
      
      while (succeeded)
      {
         pos = StringFindAnyOfChar(m_Buffer, separators);
         if (pos == 0)
         {
            ushort chr = StringGetCharacter(m_Buffer, pos);
            if (chr == '\r') needFixEOL = true;

            m_Buffer = StringSubstr(m_Buffer, 1);
            m_Pos += 1;
            break;
         }
         
         if (pos == -1)
         {
            strValue = m_Buffer;
            m_Pos += StringLen(m_Buffer);
            m_Buffer = "";
         }
         else
         {
            ushort chr = StringGetCharacter(m_Buffer, pos);
            if (chr == '\r') needFixEOL = true;

            strValue = StringSubstr(m_Buffer, 0, pos);
            m_Pos += (StringLen(strValue) + 1);
            m_Buffer = StringSubstr(m_Buffer, pos + 1);
            break;
         }

         succeeded = ReadBlock();
      }
   }
   else
   {
      ushort chr = StringGetCharacter(m_Buffer, pos);
      if (chr == '\r') needFixEOL = true;

      strValue = StringSubstr(m_Buffer, 0, pos);
      m_Pos += (StringLen(strValue) + 1);
      m_Buffer = StringSubstr(m_Buffer, pos + 1);
   }
   
   if (needFixEOL) FixEndOfLine();
   
   value = strValue;
   
   return true;
}

bool ConstFileImpl::Read(datetime& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = datetime(StringToInteger(tmp));
      return true;
   }
   return false;
}

bool ConstFileImpl::Read(color& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = StringToColor(tmp);
      return true;
   }
   return false;
}

bool ConstFileImpl::SetPos(long pos, MovePointerMethod from)
{__TRACE
   ulong newPos = 0;
   
   switch (from)
   {
      case MPM_BEGIN:
         ASSERT(pos >= 0 && pos <= long(m_Size));
         if (pos < 0 || pos > long(m_Size)) return false;
         newPos = pos;
         break;
         
      case MPM_CURRENT:
         ASSERT(pos + m_Pos >= 0 && ulong(pos + m_Pos) <= m_Size);
         if (long(pos + m_Pos) < 0 || ulong(pos + m_Pos) > m_Size) return false;
         newPos = pos + m_Pos;
         break;
         
      case MPM_END:
         ASSERT(m_Size + pos >= 0 && pos <= 0);
         if (long(m_Size + pos) < 0 || pos > 0) return false;
         newPos = m_Size + pos;
         break;
         
      default:
         ASSERT(false);
         return false;
   }
   
   m_Pos = newPos;
   m_Buffer = "";
   
   return true;
}

bool ConstFileImpl::ReadBlock(uint size)
{__TRACE
   if (m_Pos >= m_Size)
   {
      m_Buffer = "";
      return false;
   }
   
   int lo = int (m_Pos & 0xFFFFFFFF);
   int hi[1];
   hi[0] = int ((m_Pos >> 32) & 0xFFFFFFFF);
   
   ulong newPos = SetFilePointer(m_hFile, lo, hi, MPM_BEGIN);
   if (newPos == INVALID_SET_FILE_POINTER)
   {
      ASSERT(false);
      return false;
   }
   newPos = newPos + (ulong(hi[0] << 32));

   if (newPos != m_Pos)
   {
      ASSERT(false);
      return false;
   }
   
   if (m_Pos >= m_Size - 1)
   {
      m_Buffer = "";
      return false;
   }
   
   int readBytes[] = {0};
   uint toRead = size;
   if (m_Pos + toRead > m_Size)
   {
      toRead = uint(m_Size - m_Pos);
   }

   uchar buffer[];
   ArrayResize(buffer, toRead);
   
   if (ReadFile(m_hFile, buffer, toRead, readBytes, NULL))
   {
      uint newBufSize = readBytes[0];
      m_Buffer = CharArrayToStringSafe(buffer);
   }
   else ASSERT(false);
   
   return true;
}

string ConstFileImpl::ToString()
{__TRACE
   m_Buffer = "";
   
   int hi[] = {0};

   uint newPos = SetFilePointer(m_hFile, 0, hi, MPM_BEGIN);

   if (newPos == INVALID_SET_FILE_POINTER)
   {
      ASSERT(false);
      return "";
   }
   
   int readBytes[] = {0};

   uchar buffer[];
   ArrayResize(buffer, int(m_Size));
   
   if (ReadFile(m_hFile, buffer, int(m_Size), readBytes, NULL))
   {
      uint newBufSize = readBytes[0];
      
      string result = CharArrayToStringSafe(buffer);
      return result;
   }
   else ASSERT(false);
   
   return "";
}

void ConstFileImpl::FixEndOfLine(void)
{__TRACE
   if (StringLen(m_Buffer) == 0)
   {
      if (!ReadBlock()) return;
   }
   
   ushort chr = StringGetCharacter(m_Buffer, 0);
   if (chr == '\n')
   {
      m_Pos += 1;
      m_Buffer = StringSubstr(m_Buffer, 1);
   }
}

long ConstFileImpl::Find(string what)
{__TRACE
   long result = -1;
   
   ulong oldPos = m_Pos;
   
   m_Buffer = "";
   m_Pos = 0;
   
   uint length = StringLen(what);
   uint size = length + 1000;
   
   bool succeeded = ReadBlock(size);
   
   while (succeeded)
   {
      result = StringFind(m_Buffer, what);
      
      if (result != -1) break;
      
      m_Pos += length + 1000;
      m_Buffer = "";
      succeeded = ReadBlock(size);
   }
   
   m_Pos = oldPos;
   m_Buffer = "";
   
   return result;
}

bool ConstFileImpl::IsEnding() const
{
   return m_Pos >= m_Size;
}

ConstFile* OpenConstFile(string path, string separators = " ;\r\n\t") export
{__TRACE
   ConstFileImpl* result = new ConstFileImpl(path, separators);
   
   if (!result.IsValid())
   {
      ASSERT(false);
      delete result;
      result = NULL;
   }
   
   return result;
}